/*
 *
 * MariaDB Client for Java
 *
 * Copyright (c) 2012-2014 Monty Program Ab.
 * Copyright (c) 2015-2017 MariaDB Ab.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to Monty Program Ab info@montyprogram.com.
 *
 * This particular MariaDB Client for Java file is work
 * derived from a Drizzle-JDBC. Drizzle-JDBC file which is covered by subject to
 * the following copyright and notice provisions:
 *
 * Copyright (c) 2009-2011, Marcus Eriksson
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of the driver nor the names of its contributors may not be
 * used to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS  AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 */

package org.mariadb.jdbc.internal.util;

public enum SqlStates {
  WARNING("01"),
  NO_DATA("02"),
  CONNECTION_EXCEPTION("08"),
  FEATURE_NOT_SUPPORTED("0A"),
  CARDINALITY_VIOLATION("21"),
  DATA_EXCEPTION("22"),
  CONSTRAINT_VIOLATION("23"),
  INVALID_CURSOR_STATE("24"),
  INVALID_TRANSACTION_STATE("25"),
  INVALID_AUTHORIZATION("28"),
  SQL_FUNCTION_EXCEPTION("2F"),
  TRANSACTION_ROLLBACK("40"),
  SYNTAX_ERROR_ACCESS_RULE("42"),
  INVALID_CATALOG("3D"),
  INTERRUPTED_EXCEPTION("70"),
  UNDEFINED_SQLSTATE("HY"),
  TIMEOUT_EXCEPTION("JZ"),
  DISTRIBUTED_TRANSACTION_ERROR("XA"); // is this true?

  private final String sqlStateGroup;


  SqlStates(final String stateGroup) {
    this.sqlStateGroup = stateGroup;
  }

  /**
   * Get sqlState from group.
   *
   * @param group group
   * @return sqlState
   */
  public static SqlStates fromString(final String group) {
    for (final SqlStates state : SqlStates.values()) {
      if (group.startsWith(state.sqlStateGroup)) {
        return state;
      }
    }
    return UNDEFINED_SQLSTATE;
  }

  public String getSqlState() {
    return sqlStateGroup;
  }
}
